/**
 * Copyright (C) 2012-2016  Dušan Vejnovič  <vaadin@dussan.org>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.dussan.vaadin.dtable;

import java.text.Format;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.apache.commons.lang3.ArrayUtils;
import org.dussan.vaadin.dtable.client.rpc.DTableServerRpc;
import org.dussan.vaadin.dtable.component.DTableComponent;
import org.dussan.vaadin.dtable.config.ColumnType;
import org.dussan.vaadin.dtable.config.TableConfig;
import org.dussan.vaadin.dtable.config.TableExporterConfig;
import org.dussan.vaadin.dtable.config.formatting.Formatting;
import org.dussan.vaadin.dtable.config.tablesection.alignments.ColumnAlign;
import org.dussan.vaadin.dtable.processors.FilterProcessor;
import org.dussan.vaadin.dtable.processors.FormattingProcessor;
import org.dussan.vaadin.dtable.processors.OrderProcessor;
import org.dussan.vaadin.dtable.processors.SearchProcessor;
import org.dussan.vaadin.dtable.properties.filter.Filter;
import org.dussan.vaadin.dtable.properties.order.Order;
import org.dussan.vaadin.dtable.properties.search.Search;
import org.dussan.vaadin.dtable.translations.TranslatedWords;
import org.dussan.vaadin.dtable.utils.ErrorUtil;
import org.dussan.vaadin.dtable.windows.WindowsHelper;
import org.dussan.vaadin.dtable.windows.exporter.type.ExportingColumnType;

import com.vaadin.ui.UI;

public abstract class DTable
    extends DTableComponent
{

    private static final long serialVersionUID = -4291704601254104970L;

    private boolean tableInitialized = false;

    private boolean requestedDataIsProcessedNormal = false;

    private int frozenAndVisibleTableColumnsLength = 0;

    private Map<String, String> tableData = null;

    private Map<String, String> tableConfig = null;

    private Map<String, String> tableExport = null;

    private Map<String, String> translation = null;

    private Map<String, String> archivedTableData = new HashMap<>();

    private WindowsHelper windowsHelper = null;

    private DTable()
    {
        // don't remove: it define dynamic widget's width
        setWidth( "100%" );

        tableData = new HashMap<>();
        tableConfig = new HashMap<>();
        tableExport = new HashMap<>();
        translation = new HashMap<>();
        archivedTableData = new HashMap<>();
    }

    public DTable( String name )
    {
        this();
        setTableExporterConfig( new TableExporterConfig( name ) );

        registerRpc( new DTableServerRpc()
        {
            private static final long serialVersionUID = -1173210515849753962L;

            @Override
            public void newRequest( String request )
            {
                if ( processRequest( request ) )
                {
                    prepareTableData();
                    getState( true ).setTableConfigData( tableConfig );
                    getState( true ).setTableExportData( tableExport );
                    getState( true ).setTranslationData( translation );
                    getState( true ).setTableData( tableData );
                }
            }
        } );
    }

    public abstract boolean prepareData( Order[] orders, Filter[] filters, Filter[] quickSearch );

    public abstract long getAllRows( Filter[] filters, Filter[] quickSearch );

    public abstract Object[] getRowsData( int startRow, int rows, Order[] orders, Filter[] filters,
                                          Filter[] quickSearch );

    public abstract boolean prepareExportedData( String name, Order[] orders, Filter[] filters, Filter[] quickSearch );

    public abstract long getExportedAllRows( String name, Filter[] orders, Filter[] quickSearch );

    public abstract Object[] getExportedData( String name, int startRow, int rows, Order[] orders, Filter[] filters,
                                              Filter[] quickSearch );

    public abstract void exportProcessCompleted();

    private Map<String, String> resetNewRequestedData( Map<String, String> newRequetedData, int emptySearch,
                                                       String newDataObject )
    {
        if ( !getTableProperties().isOriginTableDataEmpty() )
        {
            // table is already visible on the client side
            newRequetedData.put( "noData", "noData" );
        }
        else
        {
            // no data is send to the client yet
            newRequetedData.put( "emptyData", "emptyData" );
        }

        getTableProperties().resetFilters( newDataObject );
        newRequetedData.remove( "newFilter" );
        newRequetedData.remove( "removeFilter" );

        if ( getTableConfig().isQuickSearchClearOnNoNewDataEnabled() && "quickSearch".equals( newDataObject ) )
        {
            getTableProperties().resetQuickSearch( newDataObject );
            newRequetedData.put( "emptyQuickSearch", "emptyQuickSearch" );
        }

        if ( getTableConfig().isSearchClearOnNoNewDataEnabled() && emptySearch >= 0 )
        {
            newRequetedData.put( "emptySearch", "emptySearch" );
            getTableProperties().resetSearch( newDataObject );
        }

        return newRequetedData;
    }

    private boolean processRequest( String newData )
    {
        Map<String, String> requestedData = new HashMap<>();
        if ( newData != null && !newData.isEmpty() )
        {
            for ( String data : newData.split( ARRAY_DELIMITER ) )
            {
                String[] values = data.split( ARRAY_FIELDS_DELIMITER );
                if ( values.length == 2 )
                {
                    requestedData.put( values[0], values[1] );
                }
                else
                {
                    requestedData.put( values[0], "" );
                }
            }
        }

        if ( requestedData.isEmpty() || requestedData.containsKey( "pageReload" ) )
        {
            requestedDataIsProcessedNormal = false;
        }
        else if ( requestedData.containsKey( "transmissionError" ) || !requestedData.containsKey( "top" )
            || !requestedData.containsKey( "left" ) || !requestedData.containsKey( "width" )
            || !requestedData.containsKey( "pageSize" ) || !requestedData.containsKey( "bodyRowHeight" ) )
        {
            new ErrorUtil( getTranslation().get( getTableConfig().getLocale(), TranslatedWords.ERROR ),
                           getTranslation().get( getTableConfig().getLocale(),
                                                 TranslatedWords.WHEN_TRANSMITTING_DATA_AN_ERROR_HAS_OCCURRED ) ).show();
        }
        else if ( requestedData.containsKey( "noData" ) && getTableConfig().isNoNewDataMessageEnabled() )
        {
            new ErrorUtil( getTranslation().get( getTableConfig().getLocale(), TranslatedWords.NO_NEW_DATA ),
                           getTranslation().get( getTableConfig().getLocale(),
                                                 TranslatedWords.SIZE_OF_THE_REQUESTED_DATA_IS_ZERO ) ).show();
        }
        else if ( requestedData.containsKey( "dataError" ) )
        {
            new ErrorUtil( getTranslation().get( getTableConfig().getLocale(), TranslatedWords.DATA_ERROR ),
                           getTranslation().get( getTableConfig().getLocale(),
                                                 TranslatedWords.SOMETHING_IS_WRONG_IN_DATA_PREPARATION_PLEASE_CHECK_YOUR_DATA_SETTINGS ) ).show();
        }
        else if ( requestedData.containsKey( "export" ) || requestedData.containsKey( "columnsAction" ) )
        {
            if ( windowsHelper == null )
            {
                windowsHelper = new WindowsHelper( this );
            }
            else if ( windowsHelper.getWindow() != null && windowsHelper.getWindow()
                                                                        .getParent() != null )
            {
                UI.getCurrent()
                  .removeWindow( windowsHelper.getWindow() );
            }

            // process data
            if ( requestedData.containsKey( "export" ) )
            {
                windowsHelper.process( requestedData.get( "export" ) );
            }
            else
            {
                windowsHelper.process( requestedData.get( "columnsAction" ) );
            }

            return false;
        }
        else
        {
            // standard message elements
            boolean fetchNewData = false;
            int emptySearch = -1;
            String newDataObject = new String();
            Map<String, String> newRequestedData = new HashMap<>();

            if ( getTableConfig().isFilteredFrozenTableEnabled() || getTableConfig().isFilteredTableEnabled() )
            {
                if ( requestedData.containsKey( "newFilter" ) )
                {
                    String[] newFilter = requestedData.get( "newFilter" )
                                                      .split( "_::_" );
                    int column = Integer.parseInt( newFilter[0] );
                    int row = Integer.parseInt( newFilter[1] );
                    Filter filter = FilterProcessor.process( newFilter, getTableProperties().getFilter( column ),
                                                             getTableProperties().getOriginTableData( column, row ),
                                                             getTableProperties().getColumnHeader( column ) );
                    getTableProperties().setFilter( filter );
                    fetchNewData = true;
                    newDataObject = "newFilter";
                }
                else if ( requestedData.containsKey( "removeFilter" ) )
                {
                    int column = Integer.parseInt( requestedData.get( "removeFilter" ) );
                    getTableProperties().removeFilter( column );
                    getTableProperties().removeSearch( column );
                    fetchNewData = true;
                }
                else if ( requestedData.containsKey( "quickSearch" ) && getTableConfig().isQuickSearchEnabled() )
                {
                    getTableProperties().setQuickSearch( requestedData.get( "quickSearch" ) );
                    fetchNewData = true;
                    newDataObject = "quickSearch";
                }
                else if ( requestedData.containsKey( "search" ) && getTableConfig().isSearchEnabled() )
                {
                    String searchString = requestedData.get( "search" );
                    if ( searchString != null && !searchString.isEmpty() && !"none".equals( searchString ) )
                    {
                        String[] searchValues = searchString.split( "_::_" );
                        int column = Integer.parseInt( searchValues[0] );
                        Search search =
                            SearchProcessor.process( searchValues, getTableProperties().getColumnHeader( column ),
                                                     getTableConfig().getFormattingColumn( column ) );
                        if ( search != null )
                        {
                            if ( search.getOriginalValue() != null )
                            {
                                getTableProperties().setSearches( search );
                                fetchNewData = true;
                            }
                            else if ( search.getOriginalValue() == null )
                            {
                                // error parsing format for search value
                                String error = "searchError";
                                newRequestedData.put( "noData", error );
                                newRequestedData.put( error, toStringArrayString( new String[] {
                                    "" + search.getColumn(), search.getValue() } ) );
                                fetchNewData = false;
                            }
                        }
                        else
                        {
                            getTableProperties().removeSearch( column );
                            fetchNewData = true;
                        }
                        emptySearch = column;
                    }
                    else if ( "none".equals( searchString ) && !getTableProperties().isNoSearches() )
                    {
                        getTableProperties().removeSearch();
                        fetchNewData = true;
                    }
                    newDataObject = "search";
                }
            }

            if ( requestedData.containsKey( "order" ) )
            {
                if ( getTableConfig().isOrderedFrozenTableEnabled() || getTableConfig().isOrderedTableEnabled() )
                {
                    Object orderObject = requestedData.get( "order" );
                    if ( "none".equals( orderObject ) )
                    {
                        getTableProperties().resetOrders();
                    }
                    else
                    {
                        getTableProperties().resetOrders();
                        getTableProperties().setOrders( OrderProcessor.process( orderObject,
                                                                                getTableProperties().getCompleteColumnsHeader() ) );
                    }
                    fetchNewData = true;
                }
                else
                {
                    getTableProperties().resetOrders();
                    fetchNewData = false;
                }
            }

            if ( requestedData.containsKey( "freeze" ) )
            {
                if ( getTableConfig().isFrozenTableEnabled() )
                {
                    int column = Integer.parseInt( requestedData.get( "freeze" ) );
                    if ( getTableProperties().isFrozenColumn( column ) )
                    {
                        getTableProperties().removeFrozenColumn( column );
                        getTableProperties().calculateTableWidth( getTableConfig().getColumnsWidth() );
                        fetchNewData = true;
                    }
                    else if ( ( getTableProperties().getFrozenColumns().length
                        + getTableProperties().getHiddenFrozenColumns().length ) <= getTableConfig().getTableFrozenMaximumColumns() )
                    {
                        getTableProperties().setFrozenColumn( column );
                        getTableProperties().calculateTableWidth( getTableConfig().getColumnsWidth() );
                        fetchNewData = true;
                    }
                    else
                    {
                        new ErrorUtil( getTranslation().get( getTableConfig().getLocale(), TranslatedWords.ERROR ),
                                       getTranslation().get( getTableConfig().getLocale(),
                                                             TranslatedWords.MAXIMUM_PERMITTED_NUMBER_OF_FROZEN_COLUMNS_IS )
                                           + getTableConfig().getTableFrozenMaximumColumns() ).show();
                        fetchNewData = true;
                    }
                }
                else
                {
                    getTableProperties().resetFrozenColumns();
                    fetchNewData = false;
                }
            }

            if ( requestedData.containsKey( "hideColumn" ) )
            {
                if ( "newData".equals( requestedData.get( "hideColumn" ) ) )
                {
                    fetchNewData = getTableConfig().isNewDataWhenHideUnhideColumnEnabled();
                }
                else
                {
                    int column = Integer.parseInt( requestedData.get( "hideColumn" ) );
                    if ( getTableConfig().canColumnHide( column ) )
                    {
                        getTableProperties().setHiddenColumn( column );
                        fetchNewData = getTableConfig().isNewDataWhenHideUnhideColumnEnabled();
                    }
                }
                if ( fetchNewData )
                {
                    getTableProperties().calculateTableWidth( getTableConfig().getColumnsWidth() );
                }
            }

            if ( getTableConfig().isFilteredFrozenTableEnabled() || getTableConfig().isFilteredTableEnabled()
                || requestedData.containsKey( "order" ) )
            {
                newRequestedData.put( "newFilter", toStringArrayString( getTableProperties().getNewFilter() ) );
                newRequestedData.put( "filterHtml", getTableProperties().getFilterHtml() );
                newRequestedData.put( "search", toStringArrayString( getTableProperties().getSearchString() ) );
            }

            if ( requestedData.containsKey( "newData" ) || requestedData.containsKey( "newFilter" )
                || requestedData.containsKey( "removeFilter" ) || requestedData.containsKey( "quickSearch" )
                || requestedData.containsKey( "search" ) || requestedData.containsKey( "order" )
                || requestedData.containsKey( "freeze" ) || requestedData.containsKey( "hideColumn" ) )
            {
                if ( requestedData.containsKey( "newData" ) )
                {
                    fetchNewData = Boolean.parseBoolean( requestedData.get( "newData" ) );
                }

                Order[] orders = getTableProperties().getOrders();
                Filter[] filters = getTableProperties().getCompleteFilter();
                Filter[] quickSearch =
                    getTableProperties().getQuickSearchFormatting( getTableConfig().getFormattingColumnsArray(),
                                                                   getTableConfig().isQuickSearchFromattingEnabled(),
                                                                   getTableConfig().isFilteredFrozenTableEnabled(),
                                                                   getTableConfig().isFilteredTableEnabled() );
                if ( !newRequestedData.containsKey( "noData" )
                    && ( requestedData.containsKey( "newFilter" ) || requestedData.containsKey( "removeFilter" )
                        || requestedData.containsKey( "quickSearch" ) || requestedData.containsKey( "search" ) )
                    || ( requestedData.containsKey( "hideColumn" ) && Boolean.TRUE.equals( fetchNewData ) ) )
                {
                    prepareTableQuery( getTableExporterConfig().getBaseName(), null, filters, quickSearch, 0, 0,
                                       false );

                    long allRows = 0;
                    if ( prepareData( orders, filters, quickSearch ) )
                    {
                        allRows = getAllRows( filters, quickSearch );
                    }
                    if ( 0 < allRows )
                    {
                        fetchNewData = true;
                        getTableProperties().setAllRows( allRows );
                    }
                    else
                    {
                        newRequestedData = resetNewRequestedData( newRequestedData, emptySearch, newDataObject );
                    }
                }

                newRequestedData.put( "scrollLeft", requestedData.get( "left" ) );
                newRequestedData.put( "scrollTop", requestedData.get( "top" ) );

                if ( !newRequestedData.containsKey( "noData" ) && !newRequestedData.containsKey( "emptyData" ) )
                {
                    int left = Integer.parseInt( requestedData.get( "left" ) );
                    int top = Integer.parseInt( requestedData.get( "top" ) );
                    int width = Integer.parseInt( requestedData.get( "width" ) );
                    int bodyRowHeight = Integer.parseInt( requestedData.get( "bodyRowHeight" ) );
                    getTableProperties().checkVisibleColumns( width, left, top, bodyRowHeight,
                                                              getTableConfig().getColumnsWidth(),
                                                              getTableConfig().isAllColumnsEnabled() );
                    int[] columns = getTableProperties().getFrozenAndVisibleColumns();

                    // need to send data to the client only if:
                    // - number of visible columns are not equal columns header
                    // - table is scrolling horizontally
                    // - columns is hidden and fetch new data is enabled
                    // - columns is freezing
                    // - columns is ordered
                    if ( haveToRefreshTableHeaderData( columns.length, fetchNewData,
                                                       requestedData.containsKey( "hideColumn" ),
                                                       requestedData.containsKey( "freeze" ),
                                                       requestedData.containsKey( "order" ) ) )
                    {
                        int[] visibleHeaderColumns = getTableProperties().getFrozenAndAllVisibleColumns();
                        if ( !getTableConfig().isAllColumnsEnabled() && !getTableConfig().getHeader()
                                                                                         .isAllColumnsEnabled() )
                        {
                            visibleHeaderColumns = columns;
                        }

                        newRequestedData.put( "visibleHeaderColumns", Arrays.toString( visibleHeaderColumns ) );
                        newRequestedData.put( "columnsHeader",
                                              Arrays.toString( getTableProperties().getColumnsHeader( visibleHeaderColumns ) ) );
                        newRequestedData.putAll( getTableConfig().getData( columns, visibleHeaderColumns ) );
                        newRequestedData.putAll( getTableProperties().getData() );

                        if ( getTableConfig().getHeader()
                                             .isTooltipEnabled() )
                        {
                            newRequestedData.put( "headersTooltip",
                                                  toStringArrayString( getTableProperties().getHeadersTooltip( visibleHeaderColumns ) ) );
                        }
                    }

                    // new table data
                    int firstRow = getTableProperties().getFirstRow();
                    int rows = Integer.parseInt( requestedData.get( "pageSize" ) );
                    rows = (int) ( firstRow + rows < getTableProperties().getAllRows() ? rows
                                    : getTableProperties().getAllRows() - firstRow );
                    Map<String, String[]> newTableData =
                        getTableColumnsData( firstRow, rows, columns, getTableProperties().getOrders(), filters,
                                             quickSearch, fetchNewData );
                    if ( newTableData != null )
                    {
                        for ( Entry<String, String[]> data : newTableData.entrySet() )
                        {
                            String column = data.getKey();
                            String[] columnData = data.getValue();
                            newRequestedData.put( column, toStringArrayString( columnData ) );
                        }
                        frozenAndVisibleTableColumnsLength = columns.length;
                    }

                    if ( newTableData != null && 0 < countColumnsInTableData( columns, newRequestedData.keySet() ) )
                    {
                        newRequestedData.put( "allRowsVisible", ""
                            + ( countRowsInTableData( newTableData.get( Integer.toString( columns[0] ) ) ) == getTableProperties().getAllRows() ) );
                    }
                    else
                    {
                        newRequestedData = resetNewRequestedData( newRequestedData, emptySearch, newDataObject );
                    }
                }

                // send new data to the client
                requestedDataIsProcessedNormal = true;
                this.tableData = new HashMap<>( newRequestedData );
            }
        }

        return true;
    }

    private void checkProperties()
    {
        for ( int column = 0; column < getTableProperties().getAllColumns(); column++ )
        {
            if ( ( !getTableConfig().isOrderedFrozenTableEnabled() && getTableProperties().isFrozenColumn( column ) )
                || ( !getTableConfig().isOrderedTableEnabled() && !getTableProperties().isFrozenColumn( column ) ) )
            {
                getTableProperties().removeOrder( column );
            }
            if ( ( !getTableConfig().isFilteredFrozenTableEnabled() && getTableProperties().isFrozenColumn( column ) )
                || ( !getTableConfig().isFilteredTableEnabled() && !getTableProperties().isFrozenColumn( column ) ) )
            {
                getTableProperties().removeFilter( column );
                getTableProperties().removeSearch( column );
            }
            if ( !getTableConfig().isSearchEnabled() )
            {
                getTableProperties().removeSearch( column );
            }
        }
    }

    private void prepareTableQuery( String name, Order[] orders, Filter[] filters, Filter[] quickSearch, long startRow,
                                    int rows, boolean prepapreExportedData )
    {
        getTableQuery().unlockQuery()
                       .lockQuery();

        if ( getTableExporterConfig().getBaseName()
                                     .equals( name ) )
        {
            getTableQuery().getQuery()
                           .resetOrderBy();

            boolean enableHiddenData;
            boolean enableRemovedData;
            boolean enableOriginalValues;
            Object defaultValue;
            Order[] tempOrders;
            if ( !prepapreExportedData )
            {
                enableHiddenData = getTableConfig().isPrepareDataWithHiddenColumnsEnabled();
                enableRemovedData = getTableConfig().isPrepareDataWithRemovedColumnsEnabled();
                enableOriginalValues =
                    getTableConfig().isPrepareDataWithOriginalValuesForHiddenAndRemovedColumnsEnabled();
                defaultValue = getTableConfig().getDefaultValueForHiddenAndRemovedcolumns();
                tempOrders = getTableConfig().isPrepareDataWithOrdersEnabled() ? orders : null;
            }
            else
            {
                enableHiddenData = getTableExporterConfig().isPrepareDataWithHiddenColumnsEnabled();
                enableRemovedData = getTableExporterConfig().isPrepareDataWithRemovedColumnsEnabled();
                enableOriginalValues =
                    getTableExporterConfig().isPrepareDataWithOriginalValuesForHiddenAndRemovedColumnsEnabled();
                defaultValue = getTableExporterConfig().getDefaultValueForHiddenAndRemovedColumns();
                tempOrders = getTableExporterConfig().isPrepareDataWithOrdersEnabled() ? orders : null;
            }

            int[] hiddenOrRemovedColumns = new int[] {};
            if ( !enableHiddenData || !enableRemovedData )
            {
                hiddenOrRemovedColumns =
                    getTableProperties().getHiddenAndRemovedColumns( !enableHiddenData, !enableRemovedData );
            }
            getTableQuery().setProperties( tempOrders, filters, quickSearch,
                                           getTableConfig().getFormattingColumnsArray(), hiddenOrRemovedColumns,
                                           defaultValue, enableOriginalValues );
        }
        else
        {
            getTableQuery().setProperties( orders, filters, quickSearch, getTableConfig().getFormattingColumnsArray(),
                                           new int[] {}, 0, false );
        }

        if ( 0 <= startRow && 0 < rows )
        {
            getTableQuery().getQuery()
                           .addLimit( startRow, rows );
        }
    }

    private boolean haveToRefreshTableHeaderData( int columnsLength, boolean fetchNewData, boolean hideRequest,
                                                  boolean freezeRequest, boolean orderRequest )
    {
        return columnsLength != frozenAndVisibleTableColumnsLength || Boolean.FALSE.equals( fetchNewData )
            || ( hideRequest && Boolean.TRUE.equals( fetchNewData ) )
            || ( freezeRequest && getTableConfig().isFrozenTableEnabled() )
            || ( orderRequest
                && ( getTableConfig().isOrderedFrozenTableEnabled() || getTableConfig().isOrderedTableEnabled() ) )
            || getTableProperties().getScreenWidth() == TableConfig.DEFAULT_TABLE_WIDTH;
    }

    private int countColumnsInTableData( int[] columns, Set<String> objects )
    {
        int counter = 0;
        for ( int column : columns )
        {
            if ( objects.contains( Integer.toString( column ) ) )
            {
                counter += 1;
            }
        }
        return counter;
    }

    private int countRowsInTableData( String[] objects )
    {
        try
        {
            return objects.length;
        }
        catch ( Exception e )
        {
            Logger.getLogger( DTable.class.getName() )
                  .log( Level.INFO, e.getLocalizedMessage(), e );
        }
        return 0;
    }

    private void initTable( boolean createArchive )
    {
        // set default locale
        if ( getTableConfig().getLocale() == null )
        {
            getTableConfig().setLocale( getLocale() );
        }
        else
        {
            setLocale( getTableConfig().getLocale() );
        }

        // set needed translations
        getTableProperties().setTranslatedWordIsEmpty( getTranslation().get( getTableConfig().getLocale(),
                                                                             TranslatedWords.IS_EMPTY ) );
        getTableProperties().setTranslatedWordIsNotEmpty( getTranslation().get( getTableConfig().getLocale(),
                                                                                TranslatedWords.IS_NOT_EMPTY ) );

        if ( !createArchive )
        {
            // check frozen columns
            if ( !getTableConfig().isFrozenTableEnabled() )
            {
                getTableProperties().resetFrozenColumns();
            }

            getTableProperties().checkVisibleColumns( getTableConfig().getTableWidth(),
                                                      getTableConfig().getColumnsWidth(),
                                                      getTableConfig().isAllColumnsEnabled() );
            checkProperties();

            // check for null headers in order objects
            Order[] orders = getTableProperties().getOrders();
            orders = OrderProcessor.process( orders, getTableProperties().getCompleteColumnsHeader() );
            getTableProperties().resetOrders();
            getTableProperties().setOrders( orders );

            Filter[] filters = getTableProperties().getCompleteFilter();
            Filter[] quickSearch =
                getTableProperties().getQuickSearchFormatting( getTableConfig().getFormattingColumnsArray(),
                                                               getTableConfig().isQuickSearchFromattingEnabled(),
                                                               getTableConfig().isFilteredFrozenTableEnabled(),
                                                               getTableConfig().isFilteredTableEnabled() );

            prepareTableQuery( getTableExporterConfig().getBaseName(), orders, filters, quickSearch, 0,
                               getTableConfig().getPageSize(), false );

            if ( prepareData( orders, filters, quickSearch ) )
            {
                getTableProperties().setAllRows( getAllRows( filters, quickSearch ) );
                getTableProperties().calculateTableWidth( getTableConfig().getColumnsWidth() );
            }
            else
            {
                getTableProperties().setAllRows( 0 );
            }

            if ( filters != null && 0 < filters.length
                && ( getTableConfig().isFilteredFrozenTableEnabled() || getTableConfig().isFilteredTableEnabled() ) )
            {
                tableData.put( "newFilter", toStringArrayString( getTableProperties().getNewFilter() ) );
                tableData.put( "filterHtml", getTableProperties().getFilterHtml() );
            }

            int[] columns = getTableProperties().getFrozenAndVisibleColumns();
            int[] visibleHeaderColumns = getTableProperties().getFrozenAndAllVisibleColumns();
            if ( !getTableConfig().isAllColumnsEnabled() && !getTableConfig().getHeader()
                                                                             .isAllColumnsEnabled() )
            {
                visibleHeaderColumns = columns;
            }
            tableData.put( "visibleHeaderColumns", Arrays.toString( visibleHeaderColumns ) );
            tableData.put( "columnsHeader",
                           Arrays.toString( getTableProperties().getColumnsHeader( visibleHeaderColumns ) ) );
            if ( getTableConfig().getHeader()
                                 .isTooltipEnabled() )
            {
                tableData.put( "headersTooltip",
                               Arrays.toString( getTableProperties().getHeadersTooltip( visibleHeaderColumns ) ) );
            }

            tableData.putAll( getTableConfig().getData( columns, visibleHeaderColumns ) );
            tableData.putAll( getTableProperties().getData() );

            Map<String, String[]> newTableData =
                getTableColumnsData( getTableProperties().getFirstRow(), getTableConfig().getPageSize(), columns,
                                     getTableProperties().getOrders(), filters, quickSearch, true );
            if ( newTableData != null )
            {
                for ( Entry<String, String[]> data : newTableData.entrySet() )
                {
                    String column = data.getKey();
                    String[] columnData = data.getValue();
                    tableData.put( column, toStringArrayString( columnData ) );
                }
                frozenAndVisibleTableColumnsLength = columns.length;
            }

            if ( newTableData != null && 0 < countColumnsInTableData( columns, tableData.keySet() ) )
            {
                tableData.put( "allRowsVisible", ""
                    + ( countRowsInTableData( newTableData.get( Integer.toString( columns[0] ) ) ) == getTableProperties().getAllRows() ) );
            }
            else
            {
                tableData.put( "emptyData", "emptyData" );
            }
        }
        else
        {
            tableInitialized = true;
            archivedTableData = new HashMap<>( tableData );
        }
    }

    private Map<String, Object> prepareTableData()
    {
        Map<String, Object> newData = new HashMap<>();

        // set default locale
        if ( getTableConfig().getLocale() == null )
        {
            getTableConfig().setLocale( getLocale() );
        }
        else
        {
            setLocale( getTableConfig().getLocale() );
        }

        if ( !tableInitialized || !requestedDataIsProcessedNormal )
        {
            if ( !tableInitialized )
            {
                initTable( false );
            }
            else if ( !requestedDataIsProcessedNormal )
            {
                // page reload
                tableData = new HashMap<>( archivedTableData );
            }
            tableConfig = new HashMap<>( getTableConfig().getData() );
            tableExport = new HashMap<>( getTableExporterConfig().getData() );
            translation = new HashMap<>( getTranslation().getData( getTableConfig().getLocale() ) );
        }
        else
        {
            requestedDataIsProcessedNormal = false;
            if ( !tableData.containsKey( "refreshVisibleData" ) )
            {
                tableData.putAll( getTableProperties().getData() );
            }
        }

        // create archive of table data
        if ( !tableInitialized )
        {
            initTable( true );
        }

        return newData;
    }

    @Override
    public void attach()
    {
        super.attach();
        if ( isAttached() )
        {
            prepareTableData();
            getState( true ).setTableConfigData( tableConfig );
            getState( true ).setTableExportData( tableExport );
            getState( true ).setTranslationData( translation );
            getState( true ).setTableData( tableData );
        }
    }

    public void refreshLayout( int pageRefresh )
    {
        tableData.put( "pageRefresh", "" + ( 0 < pageRefresh ? pageRefresh : TableConfig.DEFAULT_PAGE_REFRESH ) );
        getState( true ).setTableData( tableData );
    }

    public void refreshVisibleData( String cause, boolean showDialog )
    {
        requestedDataIsProcessedNormal = true;
        tableData.put( "refreshVisibleData", Boolean.toString( showDialog ) );
        tableData.put( "refreshVisibleDataCause", cause );
        getState( true ).setTableData( tableData );
    }

    public boolean prepareExportedData( String name )
    {
        Order[] orders = getTableProperties().getOrders();
        Filter[] filters = getTableProperties().getCompleteFilter();
        Filter[] quickSearch =
            getTableProperties().getQuickSearchFormatting( getTableConfig().getFormattingColumnsArray(),
                                                           getTableConfig().isQuickSearchFromattingEnabled(),
                                                           getTableConfig().isFilteredFrozenTableEnabled(),
                                                           getTableConfig().isFilteredTableEnabled() );
        return prepareExportedData( name, orders, filters, quickSearch );
    }

    public long getExportedAllRows( String name )
    {
        Filter[] filters = getTableProperties().getCompleteFilter();
        Filter[] quickSearch =
            getTableProperties().getQuickSearchFormatting( getTableConfig().getFormattingColumnsArray(),
                                                           getTableConfig().isQuickSearchFromattingEnabled(),
                                                           getTableConfig().isFilteredFrozenTableEnabled(),
                                                           getTableConfig().isFilteredTableEnabled() );

        prepareTableQuery( name, null, filters, quickSearch, 0, 0, true );
        getTableExporterConfig().setAllRows( name, getExportedAllRows( name, filters, quickSearch ) );

        return getTableExporterConfig().getAllRows( name );
    }

    public Object[] getExportedData( String name, int startRow, int rows )
    {
        Order[] orders = getTableProperties().getOrders();
        Filter[] filters = getTableProperties().getCompleteFilter();
        Filter[] quickSearch =
            getTableProperties().getQuickSearchFormatting( getTableConfig().getFormattingColumnsArray(),
                                                           getTableConfig().isQuickSearchFromattingEnabled(),
                                                           getTableConfig().isFilteredFrozenTableEnabled(),
                                                           getTableConfig().isFilteredTableEnabled() );

        prepareTableQuery( name, orders, filters, quickSearch, startRow, rows, true );

        int[] removedColumns = new int[] {};
        if ( !getTableExporterConfig().isPrepareDataWithHiddenColumnsEnabled()
            || !getTableExporterConfig().isPrepareDataWithRemovedColumnsEnabled() )
        {
            removedColumns =
                getTableProperties().getHiddenAndRemovedColumns( !getTableExporterConfig().isPrepareDataWithHiddenColumnsEnabled(),
                                                                 !getTableExporterConfig().isPrepareDataWithRemovedColumnsEnabled() );
        }

        if ( name.equals( getTableExporterConfig().getBaseName() ) && 0 < removedColumns.length )
        {
            int allColumns = getTableProperties().getAllColumns();
            Object[] data = getExportedData( name, startRow, rows, orders, filters, quickSearch );

            List<Integer> removed = new ArrayList<>();
            for ( int row = 0; row < rows; row++ )
            {
                for ( int column : removedColumns )
                {
                    removed.add( column + allColumns * row );
                }
            }
            return ArrayUtils.removeAll( data, ArrayUtils.toPrimitive( removed.toArray( new Integer[] {} ) ) );
        }

        return getExportedData( name, startRow, rows, orders, filters, quickSearch );
    }

    private Map<String, String[]> getTableColumnsData( int startRow, int rows, int[] columns, Order[] orders,
                                                       Filter[] filters, Filter[] quickSearch, boolean fetchNewData )
    {
        if ( fetchNewData )
        {
            int start = startRow;
            int step = rows;
            if ( getTableConfig().isAllRowsEnabled() )
            {
                start = 0;
                step = (int) getTableProperties().getAllRows();
            }

            prepareTableQuery( getTableExporterConfig().getBaseName(), orders, filters, quickSearch, start, step,
                               false );
            Object[] data = getRowsData( start, step, orders, filters, quickSearch );

            if ( data != null && 0 < data.length )
            {
                transformTableRows2ColumnsData( data );
            }
            else
            {
                return null;
            }
        }
        return getVisibleTableColumnsData( columns );
    }

    private void transformTableRows2ColumnsData( Object[] rowsData )
    {
        getTableProperties().resetOriginTableData();

        int allRows =
            getTableProperties().getAllColumns() == 0 ? 0 : rowsData.length / getTableProperties().getAllColumns();
        for ( int column = 0; column < getTableProperties().getAllColumns(); column++ )
        {
            List<Object> columnData = new ArrayList<>();
            for ( int row = 0; row < allRows; row++ )
            {
                columnData.add( rowsData[getTableProperties().getAllColumns() * row + column] );
            }
            getTableProperties().setOriginTableData( columnData );
        }
    }

    private Map<String, String[]> getVisibleTableColumnsData( int[] columns )
    {
        Map<String, String[]> tempColumnsData = new HashMap<>();
        for ( int column : columns )
        {
            if ( column < getTableProperties().getAllColumns() )
            {
                tempColumnsData.put( Integer.toString( column ),
                                     FormattingProcessor.process( getTableConfig().getFormattingColumn( column ),
                                                                  getTableProperties().getOriginTableData( column ) ) );
            }
        }
        return tempColumnsData;
    }

    public String setImage( String url, int width, int height )
    {
        return "image_:_" + url + "_:_" + width + "_:_" + height;
    }

    public void addColumn( String columnHeader )
    {
        setColumn( getTableProperties().getAllColumns(), columnHeader, ColumnType.TEXT, null );
    }

    public void addColumn( String columnHeader, ColumnType columnType )
    {
        setColumn( getTableProperties().getAllColumns(), columnHeader, columnType, null );
    }

    public void addColumn( String columnHeader, ColumnType columnType, Format format )
    {
        setColumn( getTableProperties().getAllColumns(), columnHeader, columnType, format );
    }

    public void setColumn( int column, String columnHeader )
    {
        setColumn( column, columnHeader, ColumnType.TEXT, null );
    }

    public void setColumn( int column, String columnHeader, ColumnType columnType )
    {
        setColumn( column, columnHeader, columnType, null );
    }

    public void setColumn( int column, String columnHeader, ColumnType columnType, Format format )
    {
        // set format
        Format columnFormat = format;
        if ( columnFormat == null )
        {
            columnFormat = Formatting.getFormat( columnType );
        }

        // base tab.le settings
        int currentColumn = Math.max( 0, column );
        getTableProperties().setColumnHeader( currentColumn, columnHeader );
        getTableConfig().setAllColumns( getTableProperties().getAllColumns() );
        getTableConfig().setColumnWidth( currentColumn, getTableConfig().getColumnDefaultWidth() );
        getTableConfig().setColumnType( currentColumn, columnType );
        getTableConfig().setFormattingColumn( currentColumn, columnFormat );
        getTableExporterConfig().setColumnDefinition( getTableExporterConfig().getBaseName(), currentColumn,
                                                      columnHeader,
                                                      getTableConfig().getFormattingColumn( currentColumn ) );

        // set columns alignment
        switch ( columnType )
        {
            case BIG_DECIMAL:
            case BIG_INTEGER:
            case DOUBLE:
            case FLOAT:
            case INTEGER:
            case LONG:
                getTableConfig().getBody()
                                .setColumnAlignment( currentColumn, ColumnAlign.RIGHT );
                break;
            case BOOLEAN:
                getTableConfig().getBody()
                                .setColumnAlignment( currentColumn, ColumnAlign.CENTER );
                break;
            case DATE:
            case MEMO:
            case TEXT:
            default:
                getTableConfig().getBody()
                                .setColumnAlignment( currentColumn, ColumnAlign.LEFT );
                break;
        }
        getTableConfig().getHeader()
                        .setColumnAlignment( currentColumn, ColumnAlign.LEFT );

        // set exporting column type
        switch ( columnType )
        {
            case BIG_DECIMAL:
                getTableExporterConfig().setColumnDefinitionExportingType( getTableExporterConfig().getBaseName(),
                                                                           currentColumn,
                                                                           ExportingColumnType.BIG_DECIMAL );
                break;
            case BIG_INTEGER:
                getTableExporterConfig().setColumnDefinitionExportingType( getTableExporterConfig().getBaseName(),
                                                                           currentColumn,
                                                                           ExportingColumnType.BIG_INTEGER );
                break;
            case BOOLEAN:
                getTableExporterConfig().setColumnDefinitionExportingType( getTableExporterConfig().getBaseName(),
                                                                           currentColumn, ExportingColumnType.BOOLEAN );
                break;
            case DATE:
                getTableExporterConfig().setColumnDefinitionExportingType( getTableExporterConfig().getBaseName(),
                                                                           currentColumn, ExportingColumnType.DATE );
                break;
            case DOUBLE:
                getTableExporterConfig().setColumnDefinitionExportingType( getTableExporterConfig().getBaseName(),
                                                                           currentColumn, ExportingColumnType.DOUBLE );
                break;
            case DECIMAL_NUMBER:
            case FLOAT:
                getTableExporterConfig().setColumnDefinitionExportingType( getTableExporterConfig().getBaseName(),
                                                                           currentColumn, ExportingColumnType.FLOAT );
                break;
            case INTEGER:
            case INTEGER_NUMBER:
                getTableExporterConfig().setColumnDefinitionExportingType( getTableExporterConfig().getBaseName(),
                                                                           currentColumn, ExportingColumnType.INTEGER );
                break;
            case LONG:
                getTableExporterConfig().setColumnDefinitionExportingType( getTableExporterConfig().getBaseName(),
                                                                           currentColumn, ExportingColumnType.LONG );
                break;
            case MEMO:
                getTableExporterConfig().setColumnDefinitionExportingType( getTableExporterConfig().getBaseName(),
                                                                           currentColumn, ExportingColumnType.MEMO );
                break;
            case TIMESTAMP:
                getTableExporterConfig().setColumnDefinitionExportingType( getTableExporterConfig().getBaseName(),
                                                                           currentColumn,
                                                                           ExportingColumnType.TIMESTAMP );
                break;
            case TEXT:
            default:
                getTableExporterConfig().setColumnDefinitionExportingType( getTableExporterConfig().getBaseName(),
                                                                           currentColumn, ExportingColumnType.TEXT );
                break;
        }
    }

    @Override
    public String toString()
    {
        return getTableExporterConfig().getBaseName();
    }

}
